"""Data Preparation class to handle data processing for machine learning."""

__author__ = 'Lukas Leufen, Felix Kleinert'
__date__ = '2020-07-20'

import datetime as dt
import logging
import os
from functools import reduce
from typing import Union, List, Iterable, Tuple, Dict

import numpy as np
import pandas as pd
import xarray as xr

from mlair.configuration import check_path_and_create
from mlair import helpers
from mlair.helpers import join, statistics
from mlair.data_handler.abstract_data_handler import AbstractDataHandler

# define a more general date type for type hinting
date = Union[dt.date, dt.datetime]
str_or_list = Union[str, List[str]]
number = Union[float, int]
num_or_list = Union[number, List[number]]
data_or_none = Union[xr.DataArray, None]

# defaults
DEFAULT_STATION_TYPE = "background"
DEFAULT_NETWORK = "AIRBASE"
DEFAULT_VAR_ALL_DICT = {'o3': 'dma8eu', 'relhum': 'average_values', 'temp': 'maximum', 'u': 'average_values',
                        'v': 'average_values', 'no': 'dma8eu', 'no2': 'dma8eu', 'cloudcover': 'average_values',
                        'pblheight': 'maximum'}
DEFAULT_WINDOW_LEAD_TIME = 3
DEFAULT_WINDOW_HISTORY_SIZE = 13
DEFAULT_TIME_DIM = "datetime"
DEFAULT_TARGET_VAR = "o3"
DEFAULT_TARGET_DIM = "variables"
DEFAULT_SAMPLING = "daily"
DEFAULT_INTERPOLATION_METHOD = "linear"


class DataHandlerSingleStation(AbstractDataHandler):

    def __init__(self, station, data_path, statistics_per_var, station_type=DEFAULT_STATION_TYPE,
                 network=DEFAULT_NETWORK, sampling=DEFAULT_SAMPLING, target_dim=DEFAULT_TARGET_DIM,
                 target_var=DEFAULT_TARGET_VAR, time_dim=DEFAULT_TIME_DIM,
                 window_history_size=DEFAULT_WINDOW_HISTORY_SIZE, window_lead_time=DEFAULT_WINDOW_LEAD_TIME,
                 interpolation_limit: int = 0, interpolation_method: str = DEFAULT_INTERPOLATION_METHOD,
                 overwrite_local_data: bool = False, transformation=None, store_data_locally: bool = True,
                 min_length: int = 0, start=None, end=None, variables=None, **kwargs):
        super().__init__()  # path, station, statistics_per_var, transformation, **kwargs)
        self.station = helpers.to_list(station)
        self.path = os.path.abspath(data_path)
        self.statistics_per_var = statistics_per_var
        self.transformation = self.setup_transformation(transformation)

        self.station_type = station_type
        self.network = network
        self.sampling = sampling
        self.target_dim = target_dim
        self.target_var = target_var
        self.time_dim = time_dim
        self.window_history_size = window_history_size
        self.window_lead_time = window_lead_time

        self.interpolation_limit = interpolation_limit
        self.interpolation_method = interpolation_method

        self.overwrite_local_data = overwrite_local_data
        self.store_data_locally = store_data_locally
        self.min_length = min_length
        self.start = start
        self.end = end

        # internal
        self.data = None
        self.meta = None
        self.variables = list(statistics_per_var.keys()) if variables is None else variables
        self.history = None
        self.label = None
        self.observation = None

        # internal for transformation
        self.mean = None
        self.std = None
        self.max = None
        self.min = None
        self._transform_method = None

        # create samples
        self.setup_samples()

    def __str__(self):
        return self.station[0]

    def __len__(self):
        assert len(self.get_X()) == len(self.get_Y())
        return len(self.get_X())

    @property
    def shape(self):
        return self.data.shape, self.get_X().shape, self.get_Y().shape

    def __repr__(self):
        return f"StationPrep(station={self.station}, data_path='{self.path}', " \
               f"statistics_per_var={self.statistics_per_var}, " \
               f"station_type='{self.station_type}', network='{self.network}', " \
               f"sampling='{self.sampling}', target_dim='{self.target_dim}', target_var='{self.target_var}', " \
               f"time_dim='{self.time_dim}', window_history_size={self.window_history_size}, " \
               f"window_lead_time={self.window_lead_time}, interpolation_limit={self.interpolation_limit}, " \
               f"interpolation_method='{self.interpolation_method}', overwrite_local_data={self.overwrite_local_data}, " \
               f"transformation={self._print_transformation_as_string})"

    @property
    def _print_transformation_as_string(self):
        str_name = ''
        if self.transformation is None:
            str_name = f'{None}'
        else:
            for k, v in self.transformation.items():
                if v is not None:
                    try:
                        v_pr = f"xr.DataArray.from_dict({v.to_dict()})"
                    except AttributeError:
                        v_pr = f"'{v}'"
                    str_name += f"'{k}':{v_pr}, "
            str_name = f"{{{str_name}}}"
        return str_name

    def get_transposed_history(self) -> xr.DataArray:
        """Return history.

        :return: history with dimensions datetime, window, Stations, variables.
        """
        return self.history.transpose("datetime", "window", "Stations", "variables").copy()

    def get_transposed_label(self) -> xr.DataArray:
        """Return label.

        :return: label with dimensions datetime*, window*, Stations, variables.
        """
        return self.label.squeeze("Stations").transpose("datetime", "window").copy()

    def get_X(self, **kwargs):
        return self.get_transposed_history()

    def get_Y(self, **kwargs):
        return self.get_transposed_label()

    def get_coordinates(self):
        coords = self.meta.loc[["station_lon", "station_lat"]].astype(float)
        return coords.rename(index={"station_lon": "lon", "station_lat": "lat"}).to_dict()[str(self)]

    def call_transform(self, inverse=False):
        self.transform(dim=self.time_dim, method=self.transformation["method"],
                       mean=self.transformation['mean'], std=self.transformation["std"],
                       min_val=self.transformation["min"], max_val=self.transformation["max"],
                       inverse=inverse
                       )

    def set_transformation(self, transformation: dict):
        if self._transform_method is not None:
            self.call_transform(inverse=True)
        self.transformation = self.setup_transformation(transformation)
        self.call_transform()
        self.make_samples()

    def setup_samples(self):
        """
        Setup samples. This method prepares and creates samples X, and labels Y.
        """
        self.load_data()
        self.interpolate(dim=self.time_dim, method=self.interpolation_method, limit=self.interpolation_limit)
        if self.transformation is not None:
            self.call_transform()
        self.make_samples()

    def make_samples(self):
        self.make_history_window(self.target_dim, self.window_history_size, self.time_dim)
        self.make_labels(self.target_dim, self.target_var, self.time_dim, self.window_lead_time)
        self.make_observation(self.target_dim, self.target_var, self.time_dim)
        self.remove_nan(self.time_dim)

    def read_data_from_disk(self, source_name=""):
        """
        Load data and meta data either from local disk (preferred) or download new data by using a custom download method.

        Data is either downloaded, if no local data is available or parameter overwrite_local_data is true. In both
        cases, downloaded data is only stored locally if store_data_locally is not disabled. If this parameter is not
        set, it is assumed, that data should be saved locally.
        """
        source_name = source_name if len(source_name) == 0 else f" from {source_name}"
        check_path_and_create(self.path)
        file_name = self._set_file_name()
        meta_file = self._set_meta_file_name()
        if self.overwrite_local_data is True:
            logging.debug(f"overwrite_local_data is true, therefore reload {file_name}{source_name}")
            if os.path.exists(file_name):
                os.remove(file_name)
            if os.path.exists(meta_file):
                os.remove(meta_file)
            data, self.meta = self.download_data(file_name, meta_file)
            logging.debug(f"loaded new data{source_name}")
        else:
            try:
                logging.debug(f"try to load local data from: {file_name}")
                data = xr.open_dataarray(file_name)
                self.meta = pd.read_csv(meta_file, index_col=0)
                self.check_station_meta()
                logging.debug("loading finished")
            except FileNotFoundError as e:
                logging.debug(e)
                logging.debug(f"load new data{source_name}")
                data, self.meta = self.download_data(file_name, meta_file)
                logging.debug("loading finished")
        # create slices and check for negative concentration.
        data = self._slice_prep(data)
        self.data = self.check_for_negative_concentrations(data)

    def download_data_from_join(self, file_name: str, meta_file: str) -> [xr.DataArray, pd.DataFrame]:
        """
        Download data from TOAR database using the JOIN interface.

        Data is transformed to a xarray dataset. If class attribute store_data_locally is true, data is additionally
        stored locally using given names for file and meta file.

        :param file_name: name of file to save data to (containing full path)
        :param meta_file: name of the meta data file (also containing full path)

        :return: downloaded data and its meta data
        """
        df_all = {}
        df, meta = join.download_join(station_name=self.station, stat_var=self.statistics_per_var,
                                      station_type=self.station_type, network_name=self.network, sampling=self.sampling)
        df_all[self.station[0]] = df
        # convert df_all to xarray
        xarr = {k: xr.DataArray(v, dims=['datetime', 'variables']) for k, v in df_all.items()}
        xarr = xr.Dataset(xarr).to_array(dim='Stations')
        if self.store_data_locally is True:
            # save locally as nc/csv file
            xarr.to_netcdf(path=file_name)
            meta.to_csv(meta_file)
        return xarr, meta

    def download_data(self, file_name, meta_file):
        data, meta = self.download_data_from_join(file_name, meta_file)
        return data, meta

    def check_station_meta(self):
        """
        Search for the entries in meta data and compare the value with the requested values.

        Will raise a FileNotFoundError if the values mismatch.
        """
        if self.station_type is not None:
            check_dict = {"station_type": self.station_type, "network_name": self.network}
            for (k, v) in check_dict.items():
                if v is None:
                    continue
                if self.meta.at[k, self.station[0]] != v:
                    logging.debug(f"meta data does not agree with given request for {k}: {v} (requested) != "
                                  f"{self.meta.at[k, self.station[0]]} (local). Raise FileNotFoundError to trigger new "
                                  f"grapping from web.")
                    raise FileNotFoundError

    def check_for_negative_concentrations(self, data: xr.DataArray, minimum: int = 0) -> xr.DataArray:
        """
        Set all negative concentrations to zero.

        Names of all concentrations are extracted from https://join.fz-juelich.de/services/rest/surfacedata/
        #2.1 Parameters. Currently, this check is applied on "benzene", "ch4", "co", "ethane", "no", "no2", "nox",
        "o3", "ox", "pm1", "pm10", "pm2p5", "propane", "so2", and "toluene".

        :param data: data array containing variables to check
        :param minimum: minimum value, by default this should be 0

        :return: corrected data
        """
        chem_vars = ["benzene", "ch4", "co", "ethane", "no", "no2", "nox", "o3", "ox", "pm1", "pm10", "pm2p5",
                     "propane", "so2", "toluene"]
        used_chem_vars = list(set(chem_vars) & set(self.variables))
        data.loc[..., used_chem_vars] = data.loc[..., used_chem_vars].clip(min=minimum)
        return data

    def shift(self, dim: str, window: int) -> xr.DataArray:
        """
        Shift data multiple times to represent history (if window <= 0) or lead time (if window > 0).

        :param dim: dimension along shift is applied
        :param window: number of steps to shift (corresponds to the window length)

        :return: shifted data
        """
        start = 1
        end = 1
        if window <= 0:
            start = window
        else:
            end = window + 1
        res = []
        for w in range(start, end):
            res.append(self.data.shift({dim: -w}))
        window_array = self.create_index_array('window', range(start, end), squeeze_dim=self.target_dim)
        res = xr.concat(res, dim=window_array)
        return res

    @staticmethod
    def create_index_array(index_name: str, index_value: Iterable[int], squeeze_dim: str) -> xr.DataArray:
        """
        Create an 1D xr.DataArray with given index name and value.

        :param index_name: name of dimension
        :param index_value: values of this dimension

        :return: this array
        """
        ind = pd.DataFrame({'val': index_value}, index=index_value)
        # res = xr.Dataset.from_dataframe(ind).to_array().rename({'index': index_name}).squeeze(dim=squeez/e_dim, drop=True)
        res = xr.Dataset.from_dataframe(ind).to_array(squeeze_dim).rename({'index': index_name}).squeeze(
            dim=squeeze_dim,
            drop=True
        )
        res.name = index_name
        return res

    def _set_file_name(self):
        all_vars = sorted(self.statistics_per_var.keys())
        return os.path.join(self.path, f"{''.join(self.station)}_{'_'.join(all_vars)}.nc")

    def _set_meta_file_name(self):
        all_vars = sorted(self.statistics_per_var.keys())
        return os.path.join(self.path, f"{''.join(self.station)}_{'_'.join(all_vars)}_meta.csv")

    def interpolate(self, dim: str, method: str = 'linear', limit: int = None, use_coordinate: Union[bool, str] = True,
                    **kwargs):
        """
        Interpolate values according to different methods.

        (Copy paste from dataarray.interpolate_na)

        :param dim:
                Specifies the dimension along which to interpolate.
        :param method:
                {'linear', 'nearest', 'zero', 'slinear', 'quadratic', 'cubic',
                          'polynomial', 'barycentric', 'krog', 'pchip',
                          'spline', 'akima'}, optional
                    String indicating which method to use for interpolation:

                    - 'linear': linear interpolation (Default). Additional keyword
                      arguments are passed to ``numpy.interp``
                    - 'nearest', 'zero', 'slinear', 'quadratic', 'cubic',
                      'polynomial': are passed to ``scipy.interpolate.interp1d``. If
                      method=='polynomial', the ``order`` keyword argument must also be
                      provided.
                    - 'barycentric', 'krog', 'pchip', 'spline', and `akima`: use their
                      respective``scipy.interpolate`` classes.
        :param limit:
                    default None
                    Maximum number of consecutive NaNs to fill. Must be greater than 0
                    or None for no limit.
        :param use_coordinate:
                default True
                    Specifies which index to use as the x values in the interpolation
                    formulated as `y = f(x)`. If False, values are treated as if
                    eqaully-spaced along `dim`. If True, the IndexVariable `dim` is
                    used. If use_coordinate is a string, it specifies the name of a
                    coordinate variariable to use as the index.
        :param kwargs:

        :return: xarray.DataArray
        """
        self.data = self.data.interpolate_na(dim=dim, method=method, limit=limit, use_coordinate=use_coordinate,
                                             **kwargs)

    def make_history_window(self, dim_name_of_inputs: str, window: int, dim_name_of_shift: str) -> None:
        """
        Create a xr.DataArray containing history data.

        Shift the data window+1 times and return a xarray which has a new dimension 'window' containing the shifted
        data. This is used to represent history in the data. Results are stored in history attribute.

        :param dim_name_of_inputs: Name of dimension which contains the input variables
        :param window: number of time steps to look back in history
                Note: window will be treated as negative value. This should be in agreement with looking back on
                a time line. Nonetheless positive values are allowed but they are converted to its negative
                expression
        :param dim_name_of_shift: Dimension along shift will be applied
        """
        window = -abs(window)
        self.history = self.shift(dim_name_of_shift, window).sel({dim_name_of_inputs: self.variables})

    def make_labels(self, dim_name_of_target: str, target_var: str_or_list, dim_name_of_shift: str,
                    window: int) -> None:
        """
        Create a xr.DataArray containing labels.

        Labels are defined as the consecutive target values (t+1, ...t+n) following the current time step t. Set label
        attribute.

        :param dim_name_of_target: Name of dimension which contains the target variable
        :param target_var: Name of target variable in 'dimension'
        :param dim_name_of_shift: Name of dimension on which xarray.DataArray.shift will be applied
        :param window: lead time of label
        """
        window = abs(window)
        self.label = self.shift(dim_name_of_shift, window).sel({dim_name_of_target: target_var})

    def make_observation(self, dim_name_of_target: str, target_var: str_or_list, dim_name_of_shift: str) -> None:
        """
        Create a xr.DataArray containing observations.

        Observations are defined as value of the current time step t. Set observation attribute.

        :param dim_name_of_target: Name of dimension which contains the observation variable
        :param target_var: Name of observation variable(s) in 'dimension'
        :param dim_name_of_shift: Name of dimension on which xarray.DataArray.shift will be applied
        """
        self.observation = self.shift(dim_name_of_shift, 0).sel({dim_name_of_target: target_var})

    def remove_nan(self, dim: str) -> None:
        """
        Remove all NAs slices along dim which contain nans in history, label and observation.

        This is done to present only a full matrix to keras.fit. Update history, label, and observation attribute.

        :param dim: dimension along the remove is performed.
        """
        intersect = []
        if (self.history is not None) and (self.label is not None):
            non_nan_history = self.history.dropna(dim=dim)
            non_nan_label = self.label.dropna(dim=dim)
            non_nan_observation = self.observation.dropna(dim=dim)
            intersect = reduce(np.intersect1d, (non_nan_history.coords[dim].values, non_nan_label.coords[dim].values,
                                                non_nan_observation.coords[dim].values))

        if len(intersect) < max(self.min_length, 1):
            self.history = None
            self.label = None
            self.observation = None
        else:
            self.history = self.history.sel({dim: intersect})
            self.label = self.label.sel({dim: intersect})
            self.observation = self.observation.sel({dim: intersect})

    def _slice_prep(self, data: xr.DataArray, coord: str = 'datetime') -> xr.DataArray:
        """
        Set start and end date for slicing and execute self._slice().

        :param data: data to slice
        :param coord: name of axis to slice

        :return: sliced data
        """
        start = self.start if self.start is not None else data.coords[coord][0].values
        end = self.end if self.end is not None else data.coords[coord][-1].values
        return self._slice(data, start, end, coord)

    @staticmethod
    def _slice(data: xr.DataArray, start: Union[date, str], end: Union[date, str], coord: str) -> xr.DataArray:
        """
        Slice through a given data_item (for example select only values of 2011).

        :param data: data to slice
        :param start: start date of slice
        :param end: end date of slice
        :param coord: name of axis to slice

        :return: sliced data
        """
        return data.loc[{coord: slice(str(start), str(end))}]

    def check_for_negative_concentrations(self, data: xr.DataArray, minimum: int = 0) -> xr.DataArray:
        """
        Set all negative concentrations to zero.

        Names of all concentrations are extracted from https://join.fz-juelich.de/services/rest/surfacedata/
        #2.1 Parameters. Currently, this check is applied on "benzene", "ch4", "co", "ethane", "no", "no2", "nox",
        "o3", "ox", "pm1", "pm10", "pm2p5", "propane", "so2", and "toluene".

        :param data: data array containing variables to check
        :param minimum: minimum value, by default this should be 0

        :return: corrected data
        """
        chem_vars = ["benzene", "ch4", "co", "ethane", "no", "no2", "nox", "o3", "ox", "pm1", "pm10", "pm2p5",
                     "propane", "so2", "toluene"]
        used_chem_vars = list(set(chem_vars) & set(self.statistics_per_var.keys()))
        data.loc[..., used_chem_vars] = data.loc[..., used_chem_vars].clip(min=minimum)
        return data

    @staticmethod
    def setup_transformation(transformation: Dict):
        """
        Set up transformation by extracting all relevant information.

        Extract all information from transformation dictionary. Possible keys are method, mean, std, min, max.
        * If a transformation should be applied on base of existing values, these need to be provided in the respective
          keys "mean" and "std" (again only if required for given method).

        :param transformation: the transformation dictionary as described above.

        :return: updated transformation dictionary

        ## Transformation

        There are two different approaches (called scopes) to transform the data:
        1) `station`: transform data for each station independently (somehow like batch normalisation)
        1) `data`: transform all data of each station with shared metrics

        Transformation must be set by the `transformation` attribute. If `transformation = None` is given to `ExperimentSetup`, 
        data is not transformed at all. For all other setups, use the following dictionary structure to specify the 
        transformation.
        ```
        transformation = {"scope": <...>, 
                        "method": <...>,
                        "mean": <...>,
                        "std": <...>}
        ExperimentSetup(..., transformation=transformation, ...)
        ```

        ### scopes

        **station**: mean and std are not used

        **data**: either provide already calculated values for mean and std (if required by transformation method), or choose 
        from different calculation schemes, explained in the mean and std section.

        ### supported transformation methods
        Currently supported methods are:
        * standardise (default, if method is not given)
        * centre

        ### mean and std
        `"mean"="accurate"`: calculate the accurate values of mean and std (depending on method) by using all data. Although, 
        this method is accurate, it may take some time for the calculation. Furthermore, this could potentially lead to memory 
        issue (not explored yet, but could appear for a very big amount of data)

        `"mean"="estimate"`: estimate mean and std (depending on method). For each station, mean and std are calculated and
        afterwards aggregated using the mean value over all station-wise metrics. This method is less accurate, especially 
        regarding the std calculation but therefore much faster.

        We recommend to use the later method *estimate* because of following reasons:
        * much faster calculation
        * real accuracy of mean and std is less important, because it is "just" a transformation / scaling
        * accuracy of mean is almost as high as in the *accurate* case, because of 
        $\bar{x_{ij}} = \bar{\left(\bar{x_i}\right)_j}$. The only difference is, that in the *estimate* case, each mean is 
        equally weighted for each station independently of the actual data count of the station.
        * accuracy of std is lower for *estimate* because of $\var{x_{ij}} \ne \bar{\left(\var{x_i}\right)_j}$, but still the mean of all 
        station-wise std is a decent estimate of the true std.

        `"mean"=<value, e.g. xr.DataArray>`: If mean and std are already calculated or shall be set manually, just add the
        scaling values instead of the calculation method. For method *centre*, std can still be None, but is required for the
        *standardise* method. **Important**: Format of given values **must** match internal data format of DataPreparation 
        class: `xr.DataArray` with `dims=["variables"]` and one value for each variable.

        """
        if transformation is None:
            return
        elif not isinstance(transformation, dict):
            raise TypeError(f"`transformation' must be either `None' or dict like e.g. `{{'method': 'standardise'}},"
                            f" but transformation is of type {type(transformation)}.")
        transformation = transformation.copy()
        method = transformation.get("method", None)
        mean = transformation.get("mean", None)
        std = transformation.get("std", None)
        max_val = transformation.get("max", None)
        min_val = transformation.get("min", None)

        transformation["method"] = method
        transformation["mean"] = mean
        transformation["std"] = std
        transformation["max"] = max_val
        transformation["min"] = min_val
        return transformation

    def load_data(self):
        try:
            self.read_data_from_disk()
        except FileNotFoundError:
            self.download_data()
            self.load_data()

    def transform(self, dim: Union[str, int] = 0, method: str = 'standardise', inverse: bool = False, mean=None,
                  std=None, min_val=None, max_val=None) -> None:
        """
        Transform data according to given transformation settings.

        This function transforms a xarray.dataarray (along dim) or pandas.DataFrame (along axis) either with mean=0
        and std=1 (`method=standardise`) or centers the data with mean=0 and no change in data scale
        (`method=centre`). Furthermore, this sets an internal instance attribute for later inverse transformation. This
        method will raise an AssertionError if an internal transform method was already set ('inverse=False') or if the
        internal transform method, internal mean and internal standard deviation weren't set ('inverse=True').

        :param string/int dim: This param is not used for inverse transformation.
                | for xarray.DataArray as string: name of dimension which should be standardised
                | for pandas.DataFrame as int: axis of dimension which should be standardised
        :param method: Choose the transformation method from 'standardise' and 'centre'. 'normalise' is not implemented
                    yet. This param is not used for inverse transformation.
        :param inverse: Switch between transformation and inverse transformation.
        :param mean: Used for transformation (if required by 'method') based on external data. If 'None' the mean is
                    calculated over the data in this class instance.
        :param std: Used for transformation (if required by 'method') based on external data. If 'None' the std is
                    calculated over the data in this class instance.
        :param min_val: Used for transformation (if required by 'method') based on external data. If 'None' min_val is
                    extracted from the data in this class instance.
        :param max_val: Used for transformation (if required by 'method') based on external data. If 'None' max_val is
                    extracted from the data in this class instance.

        :return: xarray.DataArrays or pandas.DataFrames:
                #. mean: Mean of data
                #. std: Standard deviation of data
                #. data: Standardised data
        """

        def f(data):
            if method == 'standardise':
                return statistics.standardise(data, dim)
            elif method == 'centre':
                return statistics.centre(data, dim)
            elif method == 'normalise':
                # use min/max of data or given min/max
                raise NotImplementedError
            else:
                raise NotImplementedError

        def f_apply(data):
            if method == "standardise":
                return mean, std, statistics.standardise_apply(data, mean, std)
            elif method == "centre":
                return mean, None, statistics.centre_apply(data, mean)
            else:
                raise NotImplementedError

        if not inverse:
            if self._transform_method is not None:
                raise AssertionError(f"Transform method is already set. Therefore, data was already transformed with "
                                     f"{self._transform_method}. Please perform inverse transformation of data first.")
            # apply transformation on local data instance (f) if mean is None, else apply by using mean (and std) from
            # external data.
            self.mean, self.std, self.data = locals()["f" if mean is None else "f_apply"](self.data)

            # set transform method to find correct method for inverse transformation.
            self._transform_method = method
        else:
            self.inverse_transform()

    @staticmethod
    def check_inverse_transform_params(mean: data_or_none, std: data_or_none, method: str) -> None:
        """
        Support inverse_transformation method.

        Validate if all required statistics are available for given method. E.g. centering requires mean only, whereas
        normalisation requires mean and standard deviation. Will raise an AttributeError on missing requirements.

        :param mean: data with all mean values
        :param std: data with all standard deviation values
        :param method: name of transformation method
        """
        msg = ""
        if method in ['standardise', 'centre'] and mean is None:
            msg += "mean, "
        if method == 'standardise' and std is None:
            msg += "std, "
        if len(msg) > 0:
            raise AttributeError(f"Inverse transform {method} can not be executed because following is None: {msg}")

    def inverse_transform(self) -> None:
        """
        Perform inverse transformation.

        Will raise an AssertionError, if no transformation was performed before. Checks first, if all required
        statistics are available for inverse transformation. Class attributes data, mean and std are overwritten by
        new data afterwards. Thereby, mean, std, and the private transform method are set to None to indicate, that the
        current data is not transformed.
        """

        def f_inverse(data, mean, std, method_inverse):
            if method_inverse == 'standardise':
                return statistics.standardise_inverse(data, mean, std), None, None
            elif method_inverse == 'centre':
                return statistics.centre_inverse(data, mean), None, None
            elif method_inverse == 'normalise':
                raise NotImplementedError
            else:
                raise NotImplementedError

        if self._transform_method is None:
            raise AssertionError("Inverse transformation method is not set. Data cannot be inverse transformed.")
        self.check_inverse_transform_params(self.mean, self.std, self._transform_method)
        self.data, self.mean, self.std = f_inverse(self.data, self.mean, self.std, self._transform_method)
        self._transform_method = None
        # update X and Y
        self.make_samples()

    def get_transformation_information(self, variable: str = None) -> Tuple[data_or_none, data_or_none, str]:
        """
        Extract transformation statistics and method.

        Get mean and standard deviation for given variable and the transformation method if set. If a transformation
        depends only on particular statistics (e.g. only mean is required for centering), the remaining statistics are
        returned with None as fill value.

        :param variable: Variable for which the information on transformation is requested.

        :return: mean, standard deviation and transformation method
        """
        variable = self.target_var if variable is None else variable
        try:
            mean = self.mean.sel({'variables': variable}).values
        except AttributeError:
            mean = None
        try:
            std = self.std.sel({'variables': variable}).values
        except AttributeError:
            std = None
        return mean, std, self._transform_method


if __name__ == "__main__":
    # dp = AbstractDataPrep('data/', 'dummy', 'DEBW107', ['o3', 'temp'], statistics_per_var={'o3': 'dma8eu', 'temp': 'maximum'})
    # print(dp)
    statistics_per_var = {'o3': 'dma8eu', 'temp-rea-miub': 'maximum'}
    sp = DataHandlerSingleStation(data_path='/home/felix/PycharmProjects/mlt_new/data/', station='DEBY122',
                                  statistics_per_var=statistics_per_var, station_type='background',
                                  network='UBA', sampling='daily', target_dim='variables', target_var='o3',
                                  time_dim='datetime', window_history_size=7, window_lead_time=3,
                                  interpolation_limit=0
                                  )  # transformation={'method': 'standardise'})
    # sp.set_transformation({'method': 'standardise', 'mean': sp.mean+2, 'std': sp.std+1})
    sp2 = DataHandlerSingleStation(data_path='/home/felix/PycharmProjects/mlt_new/data/', station='DEBY122',
                                   statistics_per_var=statistics_per_var, station_type='background',
                                   network='UBA', sampling='daily', target_dim='variables', target_var='o3',
                                   time_dim='datetime', window_history_size=7, window_lead_time=3,
                                   transformation={'method': 'standardise'})
    sp2.transform(inverse=True)
    sp.get_X()
    sp.get_Y()
    print(len(sp))
    print(sp.shape)
    print(sp)
